// Copyright © 2010-2024 Singular Inc. All rights reserved.

#include "SingularSDKBPLibrary.h"
#include "SingularSDK.h"
#include "SingularDelegates.h"
#include "Logging/LogMacros.h"
#include "UObject/UObjectIterator.h"

#if PLATFORM_ANDROID
#include "Android/AndroidApplication.h"
#include "Android/AndroidJNI.h"
#elif PLATFORM_IOS
#include "IOS/IOSAppDelegate.h"
#import <Singular/Singular.h>
#import <Singular/SingularConfig.h>
#import <Singular/SingularLinkParams.h>
#endif

#define UNREAL_ENGINE_SDK_NAME "UnrealEngine"
#define UNREAL_ENGINE_SDK_VERSION "2.5.0"

void BroadcastOnSingularLinksResolved(FSingularLinkParams params) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnSingularLinksResolved.Broadcast(params);
    }
}

void BroadcastOnSingularShortLinksResolved(FSingularShortLinkParams params) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnSingularShortLinksResolved.Broadcast(params);
    }
}


void BroadcastDeviceAttributionCallbackHandler(FSingularDeviceAttributionCallbackHandler deviceAttributionCallbackHandler) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->DeviceAttributionCallbackHandler.Broadcast(deviceAttributionCallbackHandler);
    }
}

void BroadcastOnSingularSdidReceivedHandler(FString result) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnSingularSdidReceived.Broadcast(result);
    }
}

void BroadcastOnSingularDidSetSdidHandler(FString result) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnSingularDidSetSdid.Broadcast(result);
    }
}

#if PLATFORM_ANDROID
// FJavaWrapper JNI args mapping
// Bool     Z
// Int      I
// Double   D
// FString  Ljava/lang/String
// Void     V
// int[]    [I
// double[] [D
// HashMap  Ljava/util/HashMap
// ArrayList  Ljava/util/ArrayList
extern "C"
{
    // utility method translating Java HashMap to Unreal C++ TMap<Fstring,FString>.
    // must be forward declared, before the calling function, since this is a C implementation.
    // Adapted from android platform code (from: `/media/jni/android_media_MediaMetadataRetriever.cpp`)
    TMap<FString, FString> JavaHashMapToTMap(JNIEnv *env, jobject hashMap)
    {
        TMap<FString, FString> tmap;
        
        // Get the Map's entry Set.
        jclass mapClass = env->FindClass("java/util/Map");
        if (mapClass == NULL) { return tmap; }
        jmethodID entrySet = env->GetMethodID(mapClass, "entrySet", "()Ljava/util/Set;");
        if (entrySet == NULL) { return tmap; }
        jobject set = env->CallObjectMethod(hashMap, entrySet);
        if (set == NULL) { return tmap; }
        
        // Obtain an iterator over the Set
        jclass setClass = env->FindClass("java/util/Set");
        if (setClass == NULL) { return tmap; }
        jmethodID iterator = env->GetMethodID(setClass, "iterator", "()Ljava/util/Iterator;");
        if (iterator == NULL) { return tmap; }
        jobject iter = env->CallObjectMethod(set, iterator);
        if (iter == NULL) { return tmap; }
        
        // Get the Iterator method IDs
        jclass iteratorClass = env->FindClass("java/util/Iterator");
        if (iteratorClass == NULL) { return tmap; }
        jmethodID hasNext = env->GetMethodID(iteratorClass, "hasNext", "()Z");
        if (hasNext == NULL) { return tmap; }
        jmethodID next = env->GetMethodID(iteratorClass, "next", "()Ljava/lang/Object;");
        if (next == NULL) { return tmap; }
        
        // Get the Entry class method IDs
        jclass entryClass = env->FindClass("java/util/Map$Entry");
        if (entryClass == NULL) { return tmap; }
        jmethodID getKey = env->GetMethodID(entryClass, "getKey", "()Ljava/lang/Object;");
        if (getKey == NULL) { return tmap; }
        jmethodID getValue = env->GetMethodID(entryClass, "getValue", "()Ljava/lang/Object;");
        if (getValue == NULL) { return tmap; }
        
        // Iterate over the entry Set
        while (env->CallBooleanMethod(iter, hasNext)) {
            jobject entry = env->CallObjectMethod(iter, next);
            jstring key = (jstring) env->CallObjectMethod(entry, getKey);
            
            const char* keyStr = env->GetStringUTFChars(key, NULL);
            if (!keyStr) { return tmap; } // Out of memory
            
            // The conversion to Double is required because
            // response from Device Attribution Info returns click_timestamp as very big number for example 1705566485086707. This is considered Double by the Unreal.
            // When used without conversion, the number is converted to 1.705566485086707E15. To avoid that conversion is done.
            jstring value;
            
            jclass jcDouble = env->FindClass("java/lang/Double");
            if (jcDouble == NULL) { return tmap; }

            jobject doubleValue = env->CallObjectMethod(entry, getValue);
            
            if (env->IsInstanceOf(doubleValue , jcDouble)) {
                jmethodID jmidDoubleValue = env->GetMethodID(jcDouble, "doubleValue", "()D" );
                jdouble convertedDoubleValue = env->CallDoubleMethod(doubleValue, jmidDoubleValue);
                
                jclass jcLong = env->FindClass("java/lang/Long");
                jmethodID longToStringMethodID = env->GetStaticMethodID(jcLong, "toString","(J)Ljava/lang/String;");
                value = (jstring) env->CallStaticObjectMethod(jcLong, longToStringMethodID, (long)convertedDoubleValue);
            } else {
                value = (jstring) env->CallObjectMethod(entry, getValue);
            }
            
            const char* valueStr = env->GetStringUTFChars(value, NULL);
            if (!valueStr) {  // Out of memory
                env->ReleaseStringUTFChars(key, keyStr);
                return tmap;
            }
            tmap.Add(FString(keyStr), FString(valueStr));
            env->ReleaseStringUTFChars(value, valueStr);
            env->DeleteLocalRef(value);
            
            env->DeleteLocalRef(entry);
            env->ReleaseStringUTFChars(key, keyStr);
            env->DeleteLocalRef(key);
        }
        
        return tmap;
    }

// fully qualified function name is different for ue4 vs ue5, see SDKDEV-157
#if ENGINE_MAJOR_VERSION == 4
    JNIEXPORT void JNICALL Java_com_epicgames_ue4_GameActivity_00024SingularUeLinkHandler_OnResolvedLink(JNIEnv *env, jobject obj, jobject linkParams)
#elif ENGINE_MAJOR_VERSION >= 5
    JNIEXPORT void JNICALL Java_com_epicgames_unreal_GameActivity_00024SingularUeLinkHandler_OnResolvedLink(JNIEnv *env, jobject obj, jobject linkParams)
#endif
    {
        if (env->IsSameObject(linkParams, NULL))
        {
            return;
        }

        TMap<FString, FString> params;
        FSingularLinkParams outputLinkParams;

        jclass clsSingularLinkParams = env->GetObjectClass(linkParams);

        jmethodID jGetDeeplinkID = env->GetMethodID(clsSingularLinkParams, "getDeeplink", "()Ljava/lang/String;");
        jmethodID jGetPassthroughID = env->GetMethodID(clsSingularLinkParams, "getPassthrough", "()Ljava/lang/String;");
        jmethodID jIsDeferredID = env->GetMethodID(clsSingularLinkParams, "isDeferred", "()Z");
        
        jstring jDeeplink = (jstring)env->CallObjectMethod(linkParams, jGetDeeplinkID);
        jstring jPassthrough = (jstring)env->CallObjectMethod(linkParams, jGetPassthroughID);
        bool isDeferred = (bool)env->CallBooleanMethod(linkParams, jIsDeferredID);

        if (jDeeplink)
        {
            const char* c_deeplink_string_value = env->GetStringUTFChars(jDeeplink, NULL);
            params.Add(TEXT("deeplink"), c_deeplink_string_value);
        }

        if (jPassthrough)
        {
            const char* c_passthrough_string_value = env->GetStringUTFChars(jPassthrough, NULL);
            params.Add(TEXT("passthrough"), c_passthrough_string_value);
        }

        params.Add(TEXT("isDeferred"), isDeferred ? TEXT("True") : TEXT("False"));

        outputLinkParams.SingularLinksParams = params;

        jmethodID jGetUrlParametersID = env->GetMethodID(clsSingularLinkParams, "getUrlParameters", "()Ljava/util/HashMap;");
        jobject urlParametersHashMap = (jobject)env->CallObjectMethod(linkParams, jGetUrlParametersID);
        TMap<FString, FString> urlParametersTMap = JavaHashMapToTMap(env, urlParametersHashMap);
        outputLinkParams.UrlParameters = urlParametersTMap;

        BroadcastOnSingularLinksResolved(outputLinkParams);
    }

// fully qualified function name is different for ue4 vs ue5, see SDKDEV-157
#if ENGINE_MAJOR_VERSION == 4
    JNIEXPORT void JNICALL Java_com_epicgames_ue4_GameActivity_00024SingularUeShortLinkHandler_OnResolvedShortLink(JNIEnv *env, jobject obj, jobject linkParams)
#elif ENGINE_MAJOR_VERSION >= 5
    JNIEXPORT void JNICALL Java_com_epicgames_unreal_GameActivity_00024SingularUeShortLinkHandler_OnResolvedShortLink(JNIEnv *env, jobject obj, jobject linkParams)
#endif
    {
        if (env->IsSameObject(linkParams, NULL))
        {
            return;
        }

        TMap<FString, FString> params;
        FSingularShortLinkParams singularShortLinkParams;

        jclass clsSingularLinkParams = env->GetObjectClass(linkParams);

        jmethodID jGetDataID = env->GetMethodID(clsSingularLinkParams, "getData", "()Ljava/lang/String;");
        jmethodID jGetErrorID = env->GetMethodID(clsSingularLinkParams, "getError", "()Ljava/lang/String;");

        jstring jData = (jstring)env->CallObjectMethod(linkParams, jGetDataID);
        jstring jError = (jstring)env->CallObjectMethod(linkParams, jGetErrorID);

        if (jData)
        {
            const char* c_data_string_value = env->GetStringUTFChars(jData, NULL);
            params.Add(TEXT("data"), c_data_string_value);
        }

        if (jError)
        {
            const char* c_error_string_value = env->GetStringUTFChars(jError, NULL);
            params.Add(TEXT("error"), c_error_string_value);
        }

        singularShortLinkParams.SingularShortLinksParams = params;
        BroadcastOnSingularShortLinksResolved(singularShortLinkParams);
    }

// fully qualified function name is different for ue4 vs ue5, see SDKDEV-157
#if ENGINE_MAJOR_VERSION == 4
    JNIEXPORT void JNICALL Java_com_epicgames_ue4_GameActivity_00024SingularUeDeviceAttributionHandler_DeviceAttributionCallback(JNIEnv *env, jobject obj, jobject deviceAttributionDataObject)
#elif ENGINE_MAJOR_VERSION >= 5
    JNIEXPORT void JNICALL Java_com_epicgames_unreal_GameActivity_00024SingularUeDeviceAttributionHandler_DeviceAttributionCallback(JNIEnv *env, jobject obj, jobject deviceAttributionDataObject)
#endif
    {
        if (env->IsSameObject(deviceAttributionDataObject, NULL))
        {
            return;
        }

        TMap<FString, FString> deviceAttributionData = JavaHashMapToTMap(env, deviceAttributionDataObject);

        FSingularDeviceAttributionCallbackHandler deviceAttributionCallbackHandler;
        deviceAttributionCallbackHandler.SingularDeviceAttributionData = deviceAttributionData;

        BroadcastDeviceAttributionCallbackHandler(deviceAttributionCallbackHandler);
    }

    // SDID accessor
    // fully qualified function name is different for ue4 vs ue5, see SDKDEV-157
    #if ENGINE_MAJOR_VERSION == 4
        JNIEXPORT void JNICALL Java_com_epicgames_ue4_GameActivity_00024SingularUeSdidAccessorHandler_SdidReceived(JNIEnv *env, jobject obj, jstring result)
    #elif ENGINE_MAJOR_VERSION >= 5
        JNIEXPORT void JNICALL Java_com_epicgames_unreal_GameActivity_00024SingularUeSdidAccessorHandler_SdidReceived(JNIEnv *env, jobject obj, jstring result)
    #endif
    {
        BroadcastOnSingularSdidReceivedHandler(env->GetStringUTFChars(result, NULL));
    }

    // fully qualified function name is different for ue4 vs ue5, see SDKDEV-157
    #if ENGINE_MAJOR_VERSION == 4
        JNIEXPORT void JNICALL Java_com_epicgames_ue4_GameActivity_00024SingularUeSdidAccessorHandler_DidSetSdid(JNIEnv *env, jobject obj, jstring result)
    #elif ENGINE_MAJOR_VERSION >= 5
        JNIEXPORT void JNICALL Java_com_epicgames_unreal_GameActivity_00024SingularUeSdidAccessorHandler_DidSetSdid(JNIEnv *env, jobject obj, jstring result)
    #endif
    {
        BroadcastOnSingularDidSetSdidHandler(env->GetStringUTFChars(result, NULL));
    }
}


jobject TmapToJNIMap(TMap<FString, FString> tmap)
{
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jclass mapClass = env->FindClass("java/util/HashMap");
    jmethodID mapConstructor = env->GetMethodID(mapClass, "<init>", "()V");
    jobject map = env->NewObject(mapClass, mapConstructor);
    jmethodID putMethod = env->GetMethodID(mapClass, "put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
    
    for (const TPair<FString, FString>& pair : tmap) {
        env->CallObjectMethod(map, putMethod, env->NewStringUTF(TCHAR_TO_UTF8(*pair.Key)), env->NewStringUTF(TCHAR_TO_UTF8(*pair.Value)));
    }
    
    return map;
}

jobject TArrayToJNIList(TArray<FString> tarray)
{
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jclass listClass = env->FindClass("java/util/ArrayList");
    jmethodID listConstructor = env->GetMethodID(listClass, "<init>", "()V");
    jobject list = env->NewObject(listClass, listConstructor);
    jmethodID addMethod = env->GetMethodID(listClass, "add", "(Ljava/lang/Object;)Z");
    
    for (FString fstring : tarray) {
      env->CallBooleanMethod(list, addMethod, env->NewStringUTF(TCHAR_TO_UTF8(*fstring)));
    }
        
    return list;
}

#elif PLATFORM_IOS

NSDictionary* TmapToNSDictionary(TMap<FString, FString> tmap)
{
    NSMutableDictionary* data = [NSMutableDictionary new];
    for (TPair<FString, FString> pair : tmap) {
        [data setObject:pair.Value.GetNSString() forKey:pair.Key.GetNSString()];
    }
    
    return data;
}

NSArray* TArrayToNSArray(TArray<FString> tarray)
{
    NSMutableArray* data = [NSMutableArray new];
    for (FString fstring : tarray) {
        [data addObject:fstring.GetNSString()];
    }

    return data;
}

NSData* convertHexStringToDataBytes(NSString* hexString) {
    if([hexString length] % 2 != 0) {
        return nil;
    }

    const char *chars = [hexString UTF8String];
    int index = 0, length = (int)[hexString length];

    NSMutableData *data = [NSMutableData dataWithCapacity:length / 2];
    char byteChars[3] = {'\0','\0','\0'};
    unsigned long wholeByte;

    while (index < length) {
        byteChars[0] = chars[index++];
        byteChars[1] = chars[index++];
        wholeByte = strtoul(byteChars, NULL, 16);
        [data appendBytes:&wholeByte length:1];
    }
    
    return data;
}

void BroadcastConversionValueUpdated(int conversionValue) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnConversionValueUpdated.Broadcast(conversionValue);
    }
}

void BroadcastConversionValuesUpdated(FSingularConversionValuesParams conversionValuesParams) {
    for (TObjectIterator<USingularDelegates> Itr; Itr; ++Itr) {
        Itr->OnConversionValuesUpdated.Broadcast(conversionValuesParams);
    }
}

#endif

USingularSDKBPLibrary::USingularSDKBPLibrary(const FObjectInitializer& ObjectInitializer)
: Super(ObjectInitializer)
{

}

void USingularSDKBPLibrary::configure()
{

}

bool USingularSDKBPLibrary::isInitialized = false;
TArray<FString> USingularSDKBPLibrary::singularEspDomains;

#if PLATFORM_IOS
// We save all the params as static from the init method
// Then they can be later used when opened with a deeplink from background.
FString USingularSDKBPLibrary::singularApiKey;
FString USingularSDKBPLibrary::singularSecret;
FString USingularSDKBPLibrary::singularCustomSdid;
int USingularSDKBPLibrary::singularSessionTimeout;
bool USingularSDKBPLibrary::singularSkanEnabled;
bool USingularSDKBPLibrary::singularClipboardAttribution;
bool USingularSDKBPLibrary::singularManualSkanConversionManagement;
int USingularSDKBPLibrary::singularWaitForTrackingAuthorizationWithTimeoutInterval;

bool USingularSDKBPLibrary::didRegisterToOpenUrl = false;
bool USingularSDKBPLibrary::didRegisterToWillGoToBackground = false;

void USingularSDKBPLibrary::InitializeIOS(NSDictionary* launchOptions, NSURL *url){
    NSString* key = singularApiKey.GetNSString();
    NSString* secret = singularSecret.GetNSString();
    
    SingularConfig* singularConfig = [[SingularConfig alloc] initWithApiKey:key andSecret:secret];
    singularConfig.skAdNetworkEnabled = singularSkanEnabled;
    singularConfig.clipboardAttribution = singularClipboardAttribution;
    singularConfig.manualSkanConversionManagement = singularManualSkanConversionManagement;
    singularConfig.waitForTrackingAuthorizationWithTimeoutInterval = singularWaitForTrackingAuthorizationWithTimeoutInterval;
    
    if (launchOptions) {
        singularConfig.launchOptions = launchOptions;
    } else if (url) {
        singularConfig.openUrl = url;
    }
    
    if (singularConfig.skAdNetworkEnabled) {
        singularConfig.conversionValueUpdatedCallback = ^(NSInteger conversionValue) {
            BroadcastConversionValueUpdated(conversionValue);
        };
        singularConfig.conversionValuesUpdatedCallback = ^(NSNumber* conversionValue, NSNumber * coarse, BOOL lock) {
            FSingularConversionValuesParams conversionValuesParams;
            conversionValuesParams.value = conversionValue ? [conversionValue intValue] : -1;
            conversionValuesParams.coarse = coarse ? [coarse intValue] : -1;
            conversionValuesParams.lock = lock;
            BroadcastConversionValuesUpdated(conversionValuesParams);
        };
    }
    
    singularConfig.singularLinksHandler = ^(SingularLinkParams *singularLinkParams) {
        FString deeplinkValue([singularLinkParams getDeepLink]);
        FString passthroughValue([singularLinkParams getPassthrough]);
        bool isDeferredValue([singularLinkParams isDeferred]);
        
        TMap<FString, FString> paramsTMap;
        paramsTMap.Add("deeplink", *deeplinkValue);
        paramsTMap.Add("passthrough", *passthroughValue);
        paramsTMap.Add("isDeferred", isDeferredValue ? TEXT("True") : TEXT("False"));
        
        FSingularLinkParams outputLinkParams;
        outputLinkParams.SingularLinksParams = paramsTMap;
        
        TMap<FString, FString> urlParamsTMap;
        NSDictionary *urlParams = [singularLinkParams getUrlParameters];
        for (NSString *urlParamKey in [urlParams allKeys]) {
            urlParamsTMap.Add(FString(urlParamKey),
                                    FString([urlParams objectForKey:urlParamKey]));
        }
        
        outputLinkParams.UrlParameters = urlParamsTMap;
        
        BroadcastOnSingularLinksResolved(outputLinkParams);
    };
    
    singularConfig.deviceAttributionCallback = ^(NSDictionary *deviceAttributionInfo) {
        
        TMap<FString, FString> deviceAttributionData;

        for (NSString *deviceAttributionKey in [deviceAttributionInfo allKeys]) {
            
            if ([[deviceAttributionInfo objectForKey:deviceAttributionKey] isKindOfClass:[NSNumber class]]) {
                NSString *convertedValue = [NSString stringWithFormat:@"%d",[[deviceAttributionInfo objectForKey:deviceAttributionKey] intValue]];
                deviceAttributionData.Add(FString(deviceAttributionKey),
                                            FString(convertedValue));
            }
            else {
                deviceAttributionData.Add(FString(deviceAttributionKey),
                                            FString([deviceAttributionInfo objectForKey:deviceAttributionKey]));
            }
        }
        
        FSingularDeviceAttributionCallbackHandler deviceAttributionCallbackHandler;
        deviceAttributionCallbackHandler.SingularDeviceAttributionData = deviceAttributionData;

        BroadcastDeviceAttributionCallbackHandler(deviceAttributionCallbackHandler);
    };


    if (singularEspDomains.Num() > 0) {
        singularConfig.espDomains = TArrayToNSArray(singularEspDomains);
    }
    
    // SDID
    singularConfig.customSdid = singularCustomSdid.GetNSString();
    if ([singularConfig.customSdid length] < 1)
    {
        singularConfig.customSdid = nil;
    }
    
    singularConfig.sdidReceivedHandler = ^(NSString *result) {
        BroadcastOnSingularSdidReceivedHandler(FString(result));
    };

    singularConfig.didSetSdidHandler = ^(NSString *result) {
        BroadcastOnSingularDidSetSdidHandler(FString(result));
    };
    
    [Singular setSessionTimeout:singularSessionTimeout];
    [Singular setWrapperName:@UNREAL_ENGINE_SDK_NAME andVersion:@UNREAL_ENGINE_SDK_VERSION];
    [Singular start:singularConfig];
}

void USingularSDKBPLibrary::OnOpenURL(UIApplication *application, NSURL *url, NSString *sourceApplication, id annotation){
    USingularSDKBPLibrary::InitializeIOS(nil, url);
}

void USingularSDKBPLibrary::OnWillResignActive(){
    if (!didRegisterToOpenUrl){
        // OnOpenURL is opened everytime the app is opened with a deeplink.
        // We register only here because we don't it to handle cold opens.
        // For cold opens we use the launch options in the init method.
        FIOSCoreDelegates::OnOpenURL.AddStatic(&OnOpenURL);
        didRegisterToOpenUrl = true;
    }
}

#endif

bool USingularSDKBPLibrary::Initialize(FString apiKey, FString apiSecret,
                                       int sessionTimeout,
                                       FString customUserId,
                                       bool skAdNetworkEnabled,
                                       bool manualSkanConversionManagement,
                                       int waitForTrackingAuthorizationWithTimeoutInterval,
                                       bool oaidCollection,
                                       bool enableLogging,
                                       int logLevel,
                                       bool clipboardAttribution,
                                       FString facebookAppId,
                                       FString customSdid)
{
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID setSingularWrapper = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "setSingularWrapper", "(Ljava/lang/String;Ljava/lang/String;)V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, setSingularWrapper, env->NewStringUTF(UNREAL_ENGINE_SDK_NAME), env->NewStringUTF(UNREAL_ENGINE_SDK_VERSION));
    
    jmethodID setSingularData = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "setSingularData", "(Ljava/util/Map;Ljava/util/List;)V", false);
    
    TMap<FString, FString> configValues;
    configValues.Add(TEXT("apikey"), apiKey);
    configValues.Add(TEXT("secret"), apiSecret);
    configValues.Add(TEXT("enableLogging"), enableLogging ? TEXT("true") : TEXT("false"));
    configValues.Add(TEXT("logLevel"), FString::FromInt(logLevel) );
    configValues.Add(TEXT("sessionTimeout"), FString::FromInt(sessionTimeout));
    configValues.Add(TEXT("oaidCollection"), oaidCollection ? TEXT("true") : TEXT("false"));
    
    if (customUserId.Len() > 0) {
        configValues.Add(TEXT("customUserId"), customUserId);
    }
    
    if (facebookAppId.Len() > 0) {
        configValues.Add(TEXT("facebookAppId"), facebookAppId);
    }
    
    if (customSdid.Len() > 0) {
        configValues.Add(TEXT("customSdid"), customSdid);
    }
    
    jobject espdomains = TArrayToJNIList(singularEspDomains);
    
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, setSingularData, TmapToJNIMap(configValues), espdomains);
    
#elif PLATFORM_IOS
    singularApiKey = apiKey;
    singularSecret = apiSecret;
    singularSessionTimeout = sessionTimeout;
    singularSkanEnabled = skAdNetworkEnabled;
    singularClipboardAttribution = clipboardAttribution;
    singularManualSkanConversionManagement = manualSkanConversionManagement;
    singularWaitForTrackingAuthorizationWithTimeoutInterval = waitForTrackingAuthorizationWithTimeoutInterval;
    
    if (customUserId.Len() > 0) {
        [Singular setCustomUserId:customUserId.GetNSString()];
    }
    
    if (customSdid.Len() > 0) {
        singularCustomSdid = customSdid.GetNSString();
    }
    
    // Here we read the launch options.
    // We read the Singular Links here only for cold starts.
    UIApplication* Application = [UIApplication sharedApplication];
    IOSAppDelegate* appDelegate = (IOSAppDelegate*)[Application delegate];
    NSDictionary* launchOptions = appDelegate.launchOptions;
    
    // We need to register to background in order to use Singular Links
    // when the app is already opened.
    if (!didRegisterToWillGoToBackground){
        FIOSCoreDelegates::OnWillResignActive.AddStatic(&OnWillResignActive);
        didRegisterToWillGoToBackground = true;
    }
    
    InitializeIOS(launchOptions, nil);
#endif
    
    // Unreal Engine uses execptions disabled flag we can't add try/catch
    isInitialized = true;
    
    return isInitialized;
}

void USingularSDKBPLibrary::SendEvent(FString eventName)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularEvent = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularEvent", "(Ljava/lang/String;)V", false);
    jstring name = env->NewStringUTF(TCHAR_TO_UTF8(*eventName));
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularEvent, name);
#elif PLATFORM_IOS
    [Singular event:eventName.GetNSString()];
#endif
}

void USingularSDKBPLibrary::SendEventWithArgs(FString eventName, TMap<FString, FString>args)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularEventWithArgs = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularEventWithArgs", "(Ljava/lang/String;Ljava/util/Map;)V", false);
    jstring name = env->NewStringUTF(TCHAR_TO_UTF8(*eventName));
    jobject map = TmapToJNIMap(args);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularEventWithArgs, name, map);
#elif PLATFORM_IOS
    [Singular event:eventName.GetNSString() withArgs:TmapToNSDictionary(args)];
#endif
}


void USingularSDKBPLibrary::CreateReferrerShortLink(FString baseLink, FString referrerName, FString referrerId,  TMap<FString, FString>args)
{
    if (!isInitialized) {
        return;
    }

#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID createReferrerShortLink = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "createReferrerShortLink", "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/util/Map;)V", false);
    jstring link = env->NewStringUTF(TCHAR_TO_UTF8(*baseLink));
    jstring name = env->NewStringUTF(TCHAR_TO_UTF8(*referrerName));
    jstring id = env->NewStringUTF(TCHAR_TO_UTF8(*referrerId));
    jobject map = TmapToJNIMap(args);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, createReferrerShortLink, link,  name, id, map);
#elif PLATFORM_IOS

    [Singular createReferrerShortLink:baseLink.GetNSString()
                        referrerName:referrerName.GetNSString()
                        referrerId:referrerId.GetNSString()
                passthroughParams:TmapToNSDictionary(args)
                completionHandler:^(NSString *data, NSError *error) {
                            FSingularShortLinkParams linkParams;
                            TMap<FString, FString> paramsDict;
                            FString dataValue(data);
                            FString errorValue(error.description);
                            paramsDict.Add("data", *dataValue);
                            paramsDict.Add("error", *errorValue);
                            linkParams.SingularShortLinksParams = paramsDict;
                            BroadcastOnSingularShortLinksResolved(linkParams);
                    }];
#endif
}


void USingularSDKBPLibrary::SendRevenue(FString eventName, FString currency, float amount)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularRevenue = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularRevenue", "(Ljava/lang/String;Ljava/lang/String;D)V", false);
    jstring name = env->NewStringUTF(TCHAR_TO_UTF8(*eventName));
    jstring cur = env->NewStringUTF(TCHAR_TO_UTF8(*currency));
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularRevenue, name, cur, amount);
#elif PLATFORM_IOS
    [Singular customRevenue:eventName.GetNSString() currency:currency.GetNSString() amount:amount];
#endif
}

void USingularSDKBPLibrary::SendRevenueWithArgs(FString eventName, FString currency, float amount, TMap<FString, FString> args)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularRevenueWithArgs = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularRevenueWithArgs", "(Ljava/lang/String;Ljava/lang/String;DLjava/util/Map;)V", false);
    jstring name = env->NewStringUTF(TCHAR_TO_UTF8(*eventName));
    jstring cur = env->NewStringUTF(TCHAR_TO_UTF8(*currency));
    jobject map = TmapToJNIMap(args);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularRevenueWithArgs, name, cur, amount, map);
#elif PLATFORM_IOS
    [Singular customRevenue:eventName.GetNSString() currency:currency.GetNSString() amount:amount withAttributes:TmapToNSDictionary(args)];
#endif
}

void USingularSDKBPLibrary::SetCustomUserId(FString customUserId)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularSetCustomUserId = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularSetCustomUserId", "(Ljava/lang/String;)V", false);
    jstring userId = env->NewStringUTF(TCHAR_TO_UTF8(*customUserId));
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularSetCustomUserId, userId);
#elif PLATFORM_IOS
    [Singular setCustomUserId:customUserId.GetNSString()];
#endif
}

void USingularSDKBPLibrary::UnsetCustomUserId()
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularUnsetCustomUserId = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularUnsetCustomUserId", "()V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularUnsetCustomUserId);
#elif PLATFORM_IOS
    [Singular unsetCustomUserId];
#endif
}

void USingularSDKBPLibrary::SetUninstallToken(FString token)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularSetUninstallToken = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularSetUninstallToken", "(Ljava/lang/String;)V", false);
    jstring uninstallToken = env->NewStringUTF(TCHAR_TO_UTF8(*token));
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularSetUninstallToken, uninstallToken);
#elif PLATFORM_IOS
    NSData *tokenData = convertHexStringToDataBytes(token.GetNSString());
    if (tokenData) {
        [Singular registerDeviceTokenForUninstall:tokenData];
    }
#endif
}

/* GDPR helpers */

void USingularSDKBPLibrary::TrackingOptIn()
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularTrackingOptIn = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularTrackingOptIn", "()V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularTrackingOptIn);
#elif PLATFORM_IOS
    [Singular trackingOptIn];
#endif
}

void USingularSDKBPLibrary::TrackingUnder13()
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularTrackingUnder13 = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularTrackingUnder13", "()V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularTrackingUnder13);
#elif PLATFORM_IOS
    [Singular trackingUnder13];
#endif
}

void USingularSDKBPLibrary::StopAllTracking()
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularStopAllTracking = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularStopAllTracking", "()V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularStopAllTracking);
#elif PLATFORM_IOS
    [Singular stopAllTracking];
#endif
}

void USingularSDKBPLibrary::ResumeAllTracking()
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularResumeAllTracking = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularResumeAllTracking", "()V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularResumeAllTracking);
#elif PLATFORM_IOS
    [Singular resumeAllTracking];
#endif
}

bool USingularSDKBPLibrary::IsAllTrackingStopped()
{
    if (!isInitialized) {
        return false;
    }
    
    bool isAllTrackingStopped = false;
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularIsAllTrackingStopped = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularIsAllTrackingStopped", "()Z", false);
    isAllTrackingStopped = FJavaWrapper::CallBooleanMethod(env, FJavaWrapper::GameActivityThis, singularIsAllTrackingStopped);
#elif PLATFORM_IOS
    isAllTrackingStopped = [Singular isAllTrackingStopped];
#endif
    return isAllTrackingStopped;
}

void USingularSDKBPLibrary::LimitDataSharing(bool shouldLimitDataSharing)
{
    if (!isInitialized) {
        return;
    }
    
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularLimitDataSharing = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularLimitDataSharing", "(Z)V", false);
    FJavaWrapper::CallVoidMethod(env, FJavaWrapper::GameActivityThis, singularLimitDataSharing, shouldLimitDataSharing);
#elif PLATFORM_IOS
    [Singular limitDataSharing:shouldLimitDataSharing];
#endif
}

bool USingularSDKBPLibrary::GetLimitDataSharing()
{
    if (!isInitialized) {
        return false;
    }
    
    bool limitDataSharing = false;
#if PLATFORM_ANDROID
    JNIEnv* env = FAndroidApplication::GetJavaEnv();
    jmethodID singularGetLimitDataSharing = FJavaWrapper::FindMethod(env, FJavaWrapper::GameActivityClassID, "singularGetLimitDataSharing", "()Z", false);
    limitDataSharing = FJavaWrapper::CallBooleanMethod(env, FJavaWrapper::GameActivityThis, singularGetLimitDataSharing);
#elif PLATFORM_IOS
    limitDataSharing = [Singular getLimitDataSharing];
#endif
    return limitDataSharing;
}

/* SKAN Methods */

bool USingularSDKBPLibrary::SkanUpdateConversionValue(int conversionValue)
{
#if PLATFORM_ANDROID
#elif PLATFORM_IOS
    NSInteger nsiConversionValue = (NSInteger)conversionValue;
    return [Singular skanUpdateConversionValue:nsiConversionValue];
#endif
    
    return true;
}

void USingularSDKBPLibrary::SkanUpdateConversionValues(int conversionValue, int coarse, bool lock )
{
#if PLATFORM_ANDROID
#elif PLATFORM_IOS
    NSInteger nsiConversionValue = (NSInteger)conversionValue;
    NSInteger nsiCoarse = (NSInteger)coarse;
    [Singular skanUpdateConversionValue:nsiConversionValue coarse:nsiCoarse lock:lock];
#endif
    
}

int USingularSDKBPLibrary::SkanGetConversionValue()
{
    int conversionValue = 0;
#if PLATFORM_ANDROID
#elif PLATFORM_IOS
    conversionValue = [[Singular skanGetConversionValue] intValue];
#endif
    return conversionValue;
}

void USingularSDKBPLibrary::SkanRegisterAppForAdNetworkAttribution()
{
#if PLATFORM_ANDROID
#elif PLATFORM_IOS
    [Singular skanRegisterAppForAdNetworkAttribution];
#endif
}
